import json
from pjrpc.server import dispatcher, validators, Method, ViewMixin

from tests.common import _


def test_method_registry():
    registry = dispatcher.MethodRegistry()

    context = object()

    def method1():
        pass

    def method2(ctx, param1):
        assert ctx is context
        assert param1 == 'param1'

    registry.add_methods(
        method1,
        Method(method2, name='custom_name2', context='ctx'),
    )

    assert registry['method1'].method is method1
    assert registry['method1'].context is None
    assert registry.get('custom_name2').method is method2
    assert registry.get('custom_name2').context == 'ctx'

    registry['method1'].bind(params=())()
    registry['custom_name2'].bind(params=dict(param1='param1'), context=context)()


async def test_method_registry_prefix():
    registry = dispatcher.MethodRegistry(prefix='prefix')

    context = object()

    async def method1():
        pass

    async def method2(ctx, param1):
        assert ctx is context
        assert param1 == 'param1'

    registry.add(method1)
    registry.add(method2, 'custom_name2', context='ctx')

    assert registry['prefix.method1'].method is method1
    assert registry['prefix.method1'].context is None
    assert registry.get('prefix.custom_name2').method is method2
    assert registry.get('prefix.custom_name2').context == 'ctx'

    await registry['prefix.method1'].bind(params=())()
    await registry['prefix.custom_name2'].bind(params=('param1',), context=context)()


def test_method_registry_merge():
    registry1 = dispatcher.MethodRegistry()
    registry2 = dispatcher.MethodRegistry()

    def method1():
        pass

    registry1.add(method1)

    def method2():
        pass

    registry2.add(method2)

    registry1.merge(registry2)

    assert registry1['method1'].method is method1
    assert registry1['method2'].method is method2


def test_method_registry_merge_prefix():
    registry1 = dispatcher.MethodRegistry(prefix='prefix1')
    registry2 = dispatcher.MethodRegistry(prefix='prefix2')

    def test_method():
        pass

    registry1.add(test_method, name='method1', context='ctx1')
    registry2.add(test_method, name='method2', context='ctx2')

    assert list(registry1.items()) == [('prefix1.method1', Method(test_method, 'prefix1.method1', 'ctx1'))]
    assert list(registry2.items()) == [('prefix2.method2', Method(test_method, 'prefix2.method2', 'ctx2'))]

    registry2.merge(registry1)

    assert list(registry2.items()) == [
        ('prefix2.method2', Method(test_method, 'prefix2.method2', 'ctx2')),
        ('prefix2.prefix1.method1', Method(test_method, 'prefix2.prefix1.method1', 'ctx1')),
    ]


async def test_method_registry_view():
    registry = dispatcher.MethodRegistry()
    validator = validators.BaseValidator()
    validator_args = {'arg': 'value'}

    class MethodView(ViewMixin):
        @validator.validate(**validator_args)
        def method1(self):
            pass

        @validator.validate(**validator_args)
        async def method2(self, param1):
            assert param1 == 'param1'

    registry.view(MethodView, prefix='view')

    assert registry['view.method1'].validator == validator
    assert registry['view.method1'].validator_args == validator_args
    assert registry['view.method2'].validator == validator
    assert registry['view.method2'].validator_args == validator_args


async def test_method_view_validation():
    registry = dispatcher.MethodRegistry()

    context = object()

    class MethodView(ViewMixin):

        def __init__(self, ctx):
            assert ctx is context

        def method1(self):
            pass

        async def method2(self, param1):
            assert param1 == 'param1'

    registry.view(MethodView, prefix='view', context='ctx')

    assert registry['view.method1'].view_cls is MethodView
    assert registry.get('view.method2').view_cls is MethodView

    registry['view.method1'].bind(params=(), context=context)()
    await registry['view.method2'].bind(params=('param1',), context=context)()


def test_dispatcher():
    disp = dispatcher.Dispatcher()

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
    })

    assert json.loads(disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'error': {
            'code': -32601,
            'message': 'Method not found',
            'data': "method 'method' not found",
        },
    }

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': None,
        'method': 'method',
    })

    assert disp.dispatch(request) is None

    def method1(param):
        return param

    disp.add(method1)

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method1',
        'params': ['param1'],
    })

    assert json.loads(disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'result': 'param1',
    }

    def method2():
        raise Exception('unhandled error')

    disp.add(method2)

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method2',
    })

    assert json.loads(disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'error': {
            'code': -32000,
            'message': 'Server error',
        },
    }

    request = json.dumps([
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': ['param1'],
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method1',
            'params': ['param2'],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method3',
        },
    ])

    assert json.loads(disp.dispatch(request)) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'param1',
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'error': {
                'code': -32601,
                'message': 'Method not found',
                'data': "method 'method3' not found",
            },
        },
    ]


def test_dispatcher_errors():
    disp = dispatcher.Dispatcher()

    assert json.loads(disp.dispatch('')) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32700,
            'message': 'Parse error',
            'data': _,
        },
    }

    assert json.loads(disp.dispatch('{}')) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': _,
        },
    }

    assert json.loads(disp.dispatch('[]')) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': _,
        },
    }

    request = json.dumps([
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
        },
    ])

    assert json.loads(disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': 'request id duplicates found: 1',
        },
    }

    assert json.loads(disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': 'request id duplicates found: 1',
        },
    }


async def test_async_dispatcher():
    disp = dispatcher.AsyncDispatcher()

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method',
    })

    assert json.loads(await disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'error': {
            'code': -32601,
            'message': 'Method not found',
            'data': "method 'method' not found",
        },
    }

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': None,
        'method': 'method',
    })

    assert await disp.dispatch(request) is None

    async def method1(param):
        return param

    disp.add(method1)

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method1',
        'params': ['param1'],
    })

    assert json.loads(await disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'result': 'param1',
    }

    async def method2():
        raise Exception('unhandled error')

    disp.add(method2)

    request = json.dumps({
        'jsonrpc': '2.0',
        'id': 1,
        'method': 'method2',
    })

    assert json.loads(await disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': 1,
        'error': {
            'code': -32000,
            'message': 'Server error',
        },
    }

    request = json.dumps([
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': ['param1'],
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method1',
            'params': ['param2'],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method3',
        },
    ])

    assert json.loads(await disp.dispatch(request)) == [
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'param1',
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'error': {
                'code': -32601,
                'message': 'Method not found',
                'data': "method 'method3' not found",
            },
        },
    ]


async def test_async_dispatcher_errors():
    disp = dispatcher.AsyncDispatcher()

    assert json.loads(await disp.dispatch('')) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32700,
            'message': 'Parse error',
            'data': _,
        },
    }

    assert json.loads(await disp.dispatch('{}')) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': _,
        },
    }

    request = json.dumps([
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
        },
    ])

    assert json.loads(await disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': 'request id duplicates found: 1',
        },
    }

    assert json.loads(await disp.dispatch(request)) == {
        'jsonrpc': '2.0',
        'id': None,
        'error': {
            'code': -32600,
            'message': 'Invalid Request',
            'data': 'request id duplicates found: 1',
        },
    }



import abc
import functools as ft
import json
import logging
from types import SimpleNamespace
from typing import Any, Callable, Dict, Generator, Iterable, Optional, Tuple, Type, Union

from pjrpc import common, Request, Response, BatchRequest, BatchResponse
from pjrpc.common import generators, exceptions, v20
from .tracer import Tracer

logger = logging.getLogger(__package__)


class Batch:
    """
    Batch wrapper. Implements some methods to wrap multiple JSON-RPC requests into a single batch request.

    :param client: JSON-RPC client instance
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method calls using dot notation.

        :param batch: batch wrapper
        """

        def __init__(self, batch: 'Batch'):
            self._batch = batch

        def __getattr__(self, attr: str) -> Callable:
            def wrapped(*args: Any, **kwargs: Any):
                self._batch.add(attr, *args, **kwargs)
                return self

            return wrapped

        def __call__(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            """
            Makes an RPC call.

            :param _trace_ctx: tracers request context
            """

            return self.call(_trace_ctx)

        def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
            """
            Makes an RPC call.

            :param _trace_ctx: tracers request context
            """

            return self._batch.call(_trace_ctx)

    @property
    def proxy(self) -> 'Proxy':
        """
        Batch request proxy object.
        """

        return Batch.Proxy(self)

    def __init__(self, client: Union['AbstractClient', 'AbstractAsyncClient']):
        self._client = client
        self._id_gen = client.id_gen_impl()
        self._requests = client.batch_request_class()

    def __getitem__(self, requests: Iterable[Tuple]) -> Any:
        """
        Adds requests to the batch and makes a request.

        :param requests: requests to be added to the batch
        :returns: request results as a tuple
        """

        self._requests.extend([
            self._client.request_class(
                method=method,
                params=params,
                id=next(self._id_gen),
            ) for method, *params in requests
        ])
        return self.call()

    def __call__(self, method: str, *args: Any, **kwargs: Any) -> 'Batch':
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        return self.add(method, *args, **kwargs)

    def add(self, method: str, *args: Any, **kwargs: Any) -> 'Batch':
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(self._client.request_class(method, args or kwargs, id=next(self._id_gen)))
        return self

    def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Any:
        """
        Makes a JSON-RPC request.

        :param _trace_ctx: tracers request context
        :returns: request results as a tuple
        """

        response = self.send(self._requests)

        return response.result if response is not None else None

    def send(
        self, request: BatchRequest, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> BatchResponse:
        """
        Sends a JSON-RPC batch request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :returns: response instance
        """

        return self._client._send(
            request,
            response_class=self._client.batch_response_class,
            validator=self._relate,
            _trace_ctx=_trace_ctx,
            **kwargs,
        )

    def notify(self, method: str, *args: Any, **kwargs: Any) -> 'Batch':
        """
        Adds a notification request to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(self._client.request_class(method, args or kwargs))
        return self

    def _relate(self, batch_request: BatchRequest, batch_response: BatchResponse) -> None:
        """
        Sets requests `related` field. if `strict` flag is ``True``
        checks that all requests have theirs corresponding responses

        :param batch_request: batch request
        :param batch_response: batch response
        """

        if batch_response.is_success:
            response_map = {response.id: response for response in batch_response if response.id is not None}

            for request in batch_request:
                if request.id is not None:
                    response = response_map.pop(request.id, None)
                    if response is None and self._client.strict:
                        raise exceptions.IdentityError(f"response '{request.id}' not found")
                    else:
                        response.related = request

            if response_map and self._client.strict:
                raise exceptions.IdentityError(f"unexpected response found: {response_map.keys()}")


class AsyncBatch(Batch):
    """
    Asynchronous batch wrapper. Used to make asynchronous JSON-RPC batch requests.
    """

    def __init__(self, client: 'AbstractAsyncClient'):
        super().__init__(client)

    async def call(self, _trace_ctx: SimpleNamespace = SimpleNamespace()) -> Optional[Any]:
        """
        Makes a JSON-RPC request.

        :param _trace_ctx: tracers request context
        :returns: request results as a tuple
        """

        response = await self.send(self._requests, _trace_ctx=_trace_ctx)

        return response.result if response is not None else None


class AbstractClient(abc.ABC):
    """
    Abstract JSON-RPC client.

    :param request_class: request class
    :param response_class: response class
    :param batch_request_class: batch request class
    :param batch_response_class: batch response class
    :param id_gen_impl: identifier generator
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    :param error_cls: JSON-RPC error base class
    :param strict: if ``True`` checks that a request and a response identifiers match
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param client: JSON-RPC client instance
        """

        def __init__(self, client: 'AbstractClient'):
            self._client = client

        def __getattr__(self, attr: str) -> Callable:
            return ft.partial(self._client.call, attr)

    @property
    def proxy(self) -> 'Proxy':
        """
        Clint proxy object.
        """

        return AbstractClient.Proxy(self)

    @property
    def batch(self) -> Batch:
        """
        Client batch wrapper.
        """

        return Batch(self)

    @abc.abstractmethod
    def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :returns: response text representation
        """

    def __init__(
        self,
        request_class: Type[Request] = v20.Request,
        response_class: Type[Response] = v20.Response,
        batch_request_class: Type[BatchRequest] = v20.BatchRequest,
        batch_response_class: Type[BatchResponse] = v20.BatchResponse,
        error_cls: Type[exceptions.JsonRpcError] = exceptions.JsonRpcError,
        id_gen_impl: Callable[..., Generator[Union[int, str], None, None]] = generators.sequential,
        json_loader: Callable = json.loads,
        json_dumper: Callable = json.dumps,
        json_encoder: Type[common.JSONEncoder] = common.JSONEncoder,
        json_decoder: Optional[json.JSONDecoder] = None,
        strict: bool = True,
        request_args: Optional[Dict[str, Any]] = None,
        tracers: Iterable[Tracer] = (),
    ):
        self.request_class = request_class
        self.response_class = response_class
        self.batch_request_class = batch_request_class
        self.batch_response_class = batch_response_class
        self.error_cls = error_cls
        self.json_loader = json_loader
        self.json_dumper = json_dumper
        self.json_encoder = json_encoder
        self.json_decoder = json_decoder
        self.id_gen_impl = id_gen_impl
        self.strict = strict
        self._request_args = request_args or {}
        self._tracers = tracers

    def __call__(self, method: str, *args: Any, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any) -> Any:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        return self.call(method, *args, **kwargs)

    def notify(self, method: str, *args: Any, _trace_ctx=SimpleNamespace(), **kwargs: Any):
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=None,
            method=method,
            params=args or kwargs,
        )
        return self.send(request, _trace_ctx=_trace_ctx)

    def call(
        self, method: str, *args: Any, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Optional[Response]:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = self.send(request, _trace_ctx=_trace_ctx)

        return response.result

    def send(
        self, request: Request, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Optional[Response]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :returns: response instance
        """

        return self._send(
            request,
            response_class=self.response_class,
            validator=self._relate,
            _trace_ctx=_trace_ctx,
            **kwargs,
        )

    def traced(method: Callable) -> Callable:
        @ft.wraps(method)
        def wrapper(self, request, _trace_ctx, **kwargs):
            for tracer in self._tracers:
                tracer.on_request_begin(_trace_ctx, request)

            try:
                response = method(self, request, _trace_ctx=_trace_ctx, **kwargs)
            except BaseException as e:
                for tracer in self._tracers:
                    tracer.on_error(_trace_ctx, request, e)
                raise

            for tracer in self._tracers:
                tracer.on_request_end(_trace_ctx, request, response)

            return response

        return wrapper

    @traced
    def _send(
        self,
        request: Request,
        response_class: Type[Response],
        validator: Callable,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[Response]:
        kwargs = {**self._request_args, **kwargs}
        request_text = self.json_dumper(request, cls=self.json_encoder)

        response_text = self._request(request_text, request.is_notification, **kwargs)
        if not request.is_notification:
            response = response_class.from_json(
                self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls,
            )
            validator(request, response)

        else:
            if self.strict and response_text:
                raise exceptions.BaseError("unexpected response")
            response = None

        return response

    def _relate(self, request: Request, response: Response) -> None:
        """
        Checks the the request and the response identifiers match.

        :param request: request
        :param response: response
        """

        if self.strict and response.id is not None and response.id != request.id:
            raise exceptions.IdentityError(
                f"response id doesn't match the request one: expected {request.id}, got {response.id}",
            )

        response.related = request


class AbstractAsyncClient(AbstractClient):
    """
    Abstract asynchronous JSON-RPC client.
    """

    @property
    def batch(self) -> AsyncBatch:
        """
        Client batch wrapper.
        """

        return AsyncBatch(self)

    @abc.abstractmethod
    async def _request(self, request_text: str, is_notification: bool = False, **kwargs: Any) -> Optional[str]:
        """
        Makes a JSON-RPC request.

        :param request_text: request text representation
        :param is_notification: is the request a notification
        :returns: response text representation
        """

    async def send(
        self, request: Request, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Optional[Response]:
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :param _trace_ctx: tracers request context
        :returns: response instance
        """

        return await self._send(
            request, _trace_ctx=_trace_ctx, response_class=self.response_class, validator=self._relate, **kwargs,
        )

    def traced(method: Callable) -> Callable:
        @ft.wraps(method)
        async def wrapper(self, request: Request, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any):
            for tracer in self._tracers:
                tracer.on_request_begin(_trace_ctx, request)

            try:
                response = await method(self, request, _trace_ctx=_trace_ctx, **kwargs)
            except BaseException as e:
                for tracer in self._tracers:
                    tracer.on_error(_trace_ctx, request, e)
                raise

            for tracer in self._tracers:
                tracer.on_request_end(_trace_ctx, request, response)

            return response

        return wrapper

    @traced
    async def _send(
        self,
        request: Request,
        response_class: Type[Response],
        validator: Callable,
        _trace_ctx: SimpleNamespace = SimpleNamespace(),
        **kwargs: Any,
    ) -> Optional[Response]:
        kwargs = {**self._request_args, **kwargs}
        request_text = self.json_dumper(request, cls=self.json_encoder)

        response_text = await self._request(request_text, request.is_notification, **kwargs)
        if not request.is_notification:
            response = response_class.from_json(
                self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls,
            )
            validator(request, response)

        else:
            if self.strict and response_text:
                raise exceptions.BaseError("unexpected response")
            response = None

        return response

    async def call(
        self, method: str, *args: Any, _trace_ctx: SimpleNamespace = SimpleNamespace(), **kwargs: Any,
    ) -> Any:
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :param _trace_ctx: tracers request context
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen_impl()),
            method=method,
            params=args or kwargs,
        )
        response = await self.send(request, _trace_ctx=_trace_ctx)

        return response.result



import json
from typing import Optional

import pydantic

import pjrpc.common.exceptions
import pjrpc.server.specs.extractors.pydantic
from pjrpc.server import Method
from pjrpc.server.specs import openapi as specs
from pjrpc.server.specs import extractors


def test_openapi_schema_generator(resources):
    spec = specs.OpenAPI(
        info=specs.Info(
            version='1.0.0',
            title='Test tittle',
            description='test api',
            contact=specs.Contact(name='owner', email='test@gmal.com'),
            license=specs.License(name='MIT'),
        ),
        external_docs=specs.ExternalDocumentation(url='http://external-doc.com'),
        servers=[
            specs.Server(
                url='http://test-server',
                description='test server',
            ),
        ],
        tags=[
            specs.Tag(name='test-tag'),
        ],
        security_schemes=dict(
            basicAuth=specs.SecurityScheme(
                type=specs.SecuritySchemeType.HTTP,
                scheme='basic',
            ),
        ),
        security=[
            dict(basicAuth=[]),
        ],
        schema_extractor=extractors.pydantic.PydanticSchemaExtractor(),
    )

    class Model(pydantic.BaseModel):
        field1: str
        field2: Optional[int] = 1

    class TestError(pjrpc.common.exceptions.JsonRpcError):
        code = 2001
        message = 'test error 1'

    @specs.annotate(
        errors=[
            TestError,
            specs.Error(code=2002, message='test error 2'),
        ],
        examples=[
            specs.MethodExample(
                params=dict(
                    param1=1,
                    param2={'field1': 'field', 'field2': 2},
                ),
                result=1,
                version='2.0',
                summary='example 1 summary',
                description='example 1 description',
            ),
        ],
        tags=['tag1', 'tag2'],
        summary='method1 summary',
        description='method1 description',
        external_docs=specs.ExternalDocumentation(url='http://doc.info#method1'),
        deprecated=True,
        security=[
            dict(basicAuth=[]),
        ],
    )
    def method1(ctx, param1: int, param2: Model, param3) -> int:
        pass

    def method2(param1: int, param2) -> Model:
        pass

    def method3(param1: Optional[float] = None, param2: int = 1) -> Optional[str]:
        pass

    @specs.annotate(
        params_schema={
            'param1': specs.Schema(
                schema={'type': 'number'},
                summary='param1 summary',
                description='param1 description',
            ),
        },
        result_schema=specs.Schema(
            schema={'type': 'number'},
            summary='result summary',
            description='result description',
        ),
    )
    def method4(param1: int) -> int:
        pass

    def method5(*args, **kwargs) -> None:
        pass

    def method6():
        pass

    method1 = Method(method1, 'method1', 'ctx')
    method2 = Method(method2, 'method2')
    method3 = Method(method3, 'method3')
    method4 = Method(method4, 'method4')
    method5 = Method(method5, 'method5')
    method6 = Method(method6, 'method6')

    actual_schema = spec.schema('/test/path', methods=[method1, method2, method3, method4, method5, method6])

    assert actual_schema == resources('openapi-1.json', loader=json.loads)


def test_ui():
    swagger_ui = specs.SwaggerUI()
    swagger_ui.get_index_page('/path')
    swagger_ui.get_static_folder()

    swagger_ui = specs.ReDoc()
    swagger_ui.get_index_page('/path')
    swagger_ui.get_static_folder()

    swagger_ui = specs.RapiDoc()
    swagger_ui.get_index_page('/path')
    swagger_ui.get_static_folder()



import json
from typing import Optional

import pydantic

import pjrpc.common.exceptions
import pjrpc.server.specs.extractors.docstring
from pjrpc.server import Method
from pjrpc.server.specs import openrpc as specs
from pjrpc.server.specs import extractors


def test_openapi_schema_generator(resources):
    spec = specs.OpenRPC(
        info=specs.Info(
            version='1.0.0',
            title='Test tittle',
            description='test api',
            contact=specs.Contact(name='owner', email='test@gmal.com'),
            license=specs.License(name='MIT'),
        ),
        external_docs=specs.ExternalDocumentation(url='http://external-doc.com'),
        servers=[
            specs.Server(
                url='http://test-server',
                name='test server',
                summary='test server summary',
                description='test server description',
            ),
        ],
        schema_extractor=extractors.docstring.DocstringSchemaExtractor(),
    )

    class Model(pydantic.BaseModel):
        field1: str
        field2: int

    class TestError(pjrpc.common.exceptions.JsonRpcError):
        code = 2001
        message = 'test error 1'

    @specs.annotate(
        errors=[
            TestError,
            specs.Error(code=2002, message='test error 2'),
        ],
        examples=[
            specs.MethodExample(
                name='Simple example',
                params=[
                    specs.ExampleObject(
                        name='param1',
                        value=1,
                        description='param1 description',
                    ),
                    specs.ExampleObject(
                        name='param2',
                        value={'field1': 'field', 'field2': 2},
                        description='param2 description',
                    ),
                ],
                result=specs.ExampleObject(
                    name='result',
                    value=1,
                    description='result description',
                ),
                summary='example 1 summary',
                description='example 1 description',
            ),
        ],
        tags=['tag1', 'tag2'],
        summary='method1 summary',
        description='method1 description',
        deprecated=True,
    )
    def method1(ctx, param1: int, param2: Model) -> int:
        """
        Method1.

        Description

        :param ctx: context
        :param integer param1: param1 summary.
                               description.
        :param object param2: param2 summary.
                              description.
        :return integer: result summary.
                         description.
        :raises TestError: test error
        """

    def method2(param1: int = 1, param2: Optional[float] = None) -> Optional[str]:
        """
        Method2.

        Description.

        :param integer? param1: param1 summary.
        :param number? param2: param2 summary.
        :return string: result summary.
        """

    @specs.annotate(
        params_schema=[
            specs.ContentDescriptor(
                name='param1',
                schema={'type': 'number'},
                summary='param1 summary',
                description='param1 description',
            ),
        ],
        result_schema=specs.ContentDescriptor(
            name='result',
            schema={'type': 'number'},
            summary='result summary',
            description='result description',
        ),
    )
    def method3(param1: int) -> int:
        pass

    def method4(*args, **kwargs) -> None:
        pass

    def method5():
        pass

    method1 = Method(method1, 'method1', 'ctx')
    method2 = Method(method2, 'method2')
    method3 = Method(method3, 'method3')
    method4 = Method(method4, 'method4')
    method5 = Method(method5, 'method5')

    actual_schema = spec.schema('/test/path', methods=[method1, method2, method3, method4, method5])

    assert actual_schema == resources('openrpc-1.json', loader=json.loads)



import json
from typing import NamedTuple

import pytest
import respx
from aioresponses import aioresponses

import pjrpc
from pjrpc.client.backend import aiohttp as aiohttp_backend
from pjrpc.client.backend import httpx as httpx_backend


class AioHttpMocker:

    class Request(NamedTuple):
        url: str
        content: str

    def __init__(self):
        self.mocker = aioresponses()

    def __enter__(self):
        self.mocker.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mocker.__exit__(exc_type, exc_val, exc_tb)

    def mock(self, method=None, url=None, status=None, content=None, json=None):
        self.mocker.add(method=method, url=url, status=status, body=content, payload=json, repeat=True)

    @property
    def requests(self):
        return [
            self.Request(url=str(key[1]), content=call.kwargs['data'])
            for key, calls in self.mocker.requests.items()
            for call in calls
        ]


class RespxMocker(respx.MockRouter):

    def mock(self, method=None, url=None, status=None, content=None, json=None):
        route = self.route(method=method, url=url)
        route.respond(status_code=status, content=content, json=json)

    @property
    def requests(self):
        return [request for request, response in self.calls]


@pytest.mark.parametrize(
    'Client, mocker', [
        (aiohttp_backend.Client, AioHttpMocker),
        (httpx_backend.AsyncClient, RespxMocker),
    ],
)
async def test_call(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        client = Client(test_url)

        response = await client.send(pjrpc.Request('method', (1, 2), id=1))
        assert response.id == 1
        assert response.result == 'result'

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = await client.call('method', arg1=1, arg2=2)
        assert result == 'result'

        assert mock.requests[1].url == test_url
        assert json.loads(mock.requests[1].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': {'arg1': 1, 'arg2': 2},
        }

        result = await client('method', 1, 2)
        assert result == 'result'

        assert mock.requests[2].url == test_url
        assert json.loads(mock.requests[2].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = await client.proxy.method(1, 2)
        assert result == 'result'

        assert mock.requests[3].url == test_url
        assert json.loads(mock.requests[3].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }


@pytest.mark.parametrize(
    'Client, mocker', [
        (aiohttp_backend.Client, AioHttpMocker),
        (httpx_backend.AsyncClient, RespxMocker),
    ],
)
async def test_notify(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock('POST', test_url, status=200, content=' ')

        client = Client(test_url)

        response = await client.send(pjrpc.Request('method', params=[1, 2]))
        assert response is None
        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': [1, 2],
        }

        response = await client.notify('method', a=1, b=2)
        assert response is None
        assert mock.requests[1].url == test_url
        assert json.loads(mock.requests[1].content) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': {'a': 1, 'b': 2},
        }


@pytest.mark.parametrize(
    'Client, mocker', [
        (aiohttp_backend.Client, AioHttpMocker),
        (httpx_backend.AsyncClient, RespxMocker),
    ],
)
async def test_batch(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json=[
                {
                    'jsonrpc': '2.0',
                    'id': 1,
                    'result': 'result1',
                },
                {
                    'jsonrpc': '2.0',
                    'id': 2,
                    'result': 2,
                },
            ],
        )

        client = Client(test_url)

        result = await client.batch.send(
            pjrpc.BatchRequest(
                pjrpc.Request('method1', params=[1, 2], id=1),
                pjrpc.Request('method2', params=[2, 3], id=2),
                pjrpc.Request('method3', params=[3, 4]),
            ),
        )
        assert len(result) == 2
        assert result[0].id == 1
        assert result[0].result == 'result1'
        assert result[1].id == 2
        assert result[1].result == 2

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method3',
                'params': [3, 4],
            },
        ]

        result = await client.batch[
            ('method1', 1, 2),
            ('method2', 2, 3),
        ]
        assert result == ('result1', 2)

        assert mock.requests[1].url == test_url
        assert json.loads(mock.requests[1].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = await client.batch('method1', 1, 2)('method2', 2, 3).call()
        assert result == ('result1', 2)

        assert mock.requests[2].url == test_url
        assert json.loads(mock.requests[2].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = await client.batch.proxy.method1(1, 2).method2(2, 3)()
        assert result == ('result1', 2)

        assert mock.requests[3].url == test_url
        assert json.loads(mock.requests[3].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

    with mocker() as mock:
        mock.mock('POST', test_url, status=200, content=' ')
        result = await client.batch.notify('method1', 1, 2).notify('method2', 2, 3).call()
        assert result is None

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == [
            {
                'jsonrpc': '2.0',
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method2',
                'params': [2, 3],
            },
        ]


@pytest.mark.parametrize(
    'Client, mocker', [
        (aiohttp_backend.Client, AioHttpMocker),
        (httpx_backend.AsyncClient, RespxMocker),
    ],
)
async def test_context_manager(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        async with Client(test_url) as client:
            response = await client.send(pjrpc.Request('method', (1, 2), id=1))

            assert response.id == 1
            assert response.result == 'result'



import json
import pytest

import pjrpc
from pjrpc.client.integrations.pytest import PjRpcMocker


class TestClient:
    def __init__(self, endpoint):
        self._endpoint = endpoint

    def _request(self, data, is_notification=False, **kwargs):
        return json.dumps(pjrpc.Response(id='original_id', result='original_result').to_json())


@pytest.fixture
def endpoint():
    return 'endpoint'


@pytest.fixture
def cli(endpoint):
    return TestClient(endpoint)


def test_context_manager(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request') as mocker:
        mocker.add(endpoint, 'method', result='result')
        cli._request(json.dumps(pjrpc.Request(method='method').to_json()))

    assert mocker._patcher not in mocker._patcher._active_patches
    assert not mocker._matches
    assert not mocker._calls


def test_pjrpc_mocker_result_error_id(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request') as mocker:
        mocker.add(endpoint, 'method1', result='result')
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method1').to_json())),
            ),
        )
        assert response.result == 'result'

        mocker.add(endpoint, 'method2', error=pjrpc.exc.JsonRpcError(code=1, message='message'))
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method2').to_json())),
            ),
        )

        assert response.error == pjrpc.exc.JsonRpcError(code=1, message='message')


def test_pjrpc_mocker_once_param(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request', passthrough=True) as mocker:
        mocker.add(endpoint, 'method', result='result', once=True)
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result'

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'original_result'


def test_pjrpc_mocker_round_robin(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request') as mocker:
        mocker.add(endpoint, 'method', result='result1')
        mocker.add(endpoint, 'method', result='result2')

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result1'

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result2'

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result1'


def test_pjrpc_replace_remove(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request', passthrough=True) as mocker:
        mocker.add(endpoint, 'method', result='result1')
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result1'

        mocker.replace(endpoint, 'method', result='result2')
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'result2'

        mocker.remove(endpoint, 'method')
        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method').to_json())),
            ),
        )
        assert response.result == 'original_result'


def test_pjrpc_mocker_calls(endpoint):
    cli1 = TestClient('endpoint1')
    cli2 = TestClient('endpoint2')

    with PjRpcMocker('test_pytest_plugin.TestClient._request') as mocker:
        mocker.add('endpoint1', 'method1', result='result')
        mocker.add('endpoint1', 'method2', result='result')
        mocker.add('endpoint2', 'method1', result='result')

        cli1._request(json.dumps(pjrpc.Request(method='method1', params=[1, '2']).to_json()))
        cli1._request(json.dumps(pjrpc.Request(method='method1', params=[1, '2']).to_json()))
        cli1._request(json.dumps(pjrpc.Request(method='method2', params=[1, '2']).to_json()))
        cli2._request(json.dumps(pjrpc.Request(method='method1', params={'a': 1, 'b': '2'}).to_json()))

        assert mocker.calls['endpoint1'][('2.0', 'method1')].mock_calls == [((1, '2'), {}), ((1, '2'), {})]
        assert mocker.calls['endpoint1'][('2.0', 'method2')].mock_calls == [((1, '2'), {})]
        assert mocker.calls['endpoint2'][('2.0', 'method1')].mock_calls == [({'a': 1, 'b': '2'},)]


def test_pjrpc_mocker_callback(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request') as mocker:
        def callback(**kwargs):
            assert kwargs == {'a': 1, 'b': '2'}
            return 'result'

        mocker.add(endpoint, 'method', callback=callback)

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method', params={'a': 1, 'b': '2'}).to_json())),
            ),
        )

        assert response.result == 'result'


def test_pjrpc_mocker_passthrough(cli, endpoint):
    with PjRpcMocker('test_pytest_plugin.TestClient._request', passthrough=True) as mocker:
        mocker.add('other_endpoint', 'method', result='result')

        response = pjrpc.Response.from_json(
            json.loads(
                cli._request(json.dumps(pjrpc.Request(method='method2').to_json())),
            ),
        )

        assert response.result == 'original_result'


class TestAsyncClient:
    def __init__(self, endpoint):
        self._endpoint = endpoint

    async def _request(self, data, is_notification=False, **kwargs):
        return json.dumps(pjrpc.Response(id='original_id', result='original_result').to_json())


async def test_pjrpc_mocker_async(endpoint):
    cli = TestAsyncClient(endpoint)

    with PjRpcMocker('test_pytest_plugin.TestAsyncClient._request') as mocker:
        mocker.add(endpoint, 'method1', result='result1')
        mocker.add(endpoint, 'method2', result='result2')

        batch = pjrpc.BatchResponse.from_json(
            json.loads(
                await cli._request(
                    json.dumps(
                        pjrpc.BatchRequest(
                            pjrpc.Request(method='method1'),
                            pjrpc.Request(method='method2'),
                        ).to_json(),
                    ),
                ),
            ),
        )

        assert batch[0].result == 'result1'
        assert batch[1].result == 'result2'



"""
aiohttp JSON-RPC server integration.
"""

import json
from typing import Any, Optional
from aiohttp import web

import pjrpc
from pjrpc.server import specs, utils


class Application:
    """
    `aiohttp <https://aiohttp.readthedocs.io/en/stable/web.html>`_ based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param app_args: arguments to be passed to :py:class:`aiohttp.web.Application`
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.AsyncDispatcher`
    """

    def __init__(
        self,
        path: str = '',
        spec: Optional[specs.Specification] = None,
        app: Optional[web.Application] = None,
        **kwargs: Any
    ):
        self._path = path.rstrip('/')
        self._spec = spec
        self._app = app or web.Application()
        self._dispatcher = pjrpc.server.AsyncDispatcher(**kwargs)

        self._app.router.add_post(self._path, self._rpc_handle)

        if self._spec:
            self._app.router.add_get(utils.join_path(self._path, self._spec.path), self._generate_spec)

            if self._spec.ui and self._spec.ui_path:
                ui_app = web.Application()
                ui_app.router.add_get('/', self._ui_index_page)
                ui_app.router.add_get('/index.html', self._ui_index_page)
                ui_app.router.add_static('/', self._spec.ui.get_static_folder())

                self._app.add_subapp(utils.join_path(self._path, self._spec.ui_path), ui_app)

    @property
    def app(self) -> web.Application:
        """
        aiohttp application.
        """

        return self._app

    @property
    def dispatcher(self) -> pjrpc.server.Dispatcher:
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def _generate_spec(self, request: web.Request) -> web.Response:
        spec_full_path = utils.remove_suffix(request.path, suffix=self._spec.path)
        schema = self._spec.schema(path=spec_full_path, methods=self._dispatcher.registry.values())

        return web.json_response(text=json.dumps(schema, indent=2, cls=specs.JSONEncoder))

    async def _ui_index_page(self, request: web.Request) -> web.Response:
        app_path = request.path.rsplit(self._spec.ui_path, maxsplit=1)[0]
        spec_full_path = utils.join_path(app_path, self._spec.path)

        return web.Response(
            text=self._spec.ui.get_index_page(spec_url=spec_full_path),
            content_type='text/html',
        )

    async def _rpc_handle(self, http_request: web.Request) -> web.Response:
        """
        Handles JSON-RPC request.

        :param http_request: :py:class:`aiohttp.web.Response`
        :returns: :py:class:`aiohttp.web.Request`
        """

        if http_request.content_type != 'application/json':
            raise web.HTTPUnsupportedMediaType()

        try:
            request_text = await http_request.text()
        except UnicodeDecodeError as e:
            raise web.HTTPBadRequest() from e

        response_text = await self._dispatcher.dispatch(request_text, context=http_request)
        if response_text is None:
            return web.Response()
        else:
            return web.json_response(text=response_text)



import pytest
from aiohttp import web

from pjrpc import exc
from pjrpc.common import v20
from pjrpc.server.integration import aiohttp as integration

from tests.common import _


@pytest.fixture
def path():
    return '/test/path'


@pytest.fixture
def json_rpc(path):
    json_rpc = integration.Application(path)

    return json_rpc


@pytest.mark.parametrize(
    'request_id, params, result', [
        (
                1,
                (1, 1.1, 'str', {}, False),
                [1, 1.1, 'str', {}, False],
        ),
        (
                'abc',
                {'int': 1, 'float': 1.1, 'str': 'str', 'dict': {}, 'bool': False},
                {'int': 1, 'float': 1.1, 'str': 'str', 'dict': {}, 'bool': False},
        ),
    ],
)
async def test_request(json_rpc, path, mocker, aiohttp_client, request_id, params, result):
    method_name = 'test_method'
    mock = mocker.Mock(name=method_name, return_value=result)

    json_rpc.dispatcher.add(mock, method_name)

    cli = await aiohttp_client(json_rpc.app)
    raw = await cli.post(path, json=v20.Request(method=method_name, params=params, id=request_id).to_json())
    assert raw.status == 200

    resp = v20.Response.from_json(await raw.json())

    if isinstance(params, dict):
        mock.assert_called_once_with(kwargs=params)
    else:
        mock.assert_called_once_with(args=params)

    assert resp.id == request_id
    assert resp.result == result


async def test_notify(json_rpc, path, mocker, aiohttp_client):
    params = [1, 2]
    method_name = 'test_method'
    mock = mocker.Mock(name=method_name, return_value='result')

    json_rpc.dispatcher.add(mock, method_name)

    cli = await aiohttp_client(json_rpc.app)
    raw = await cli.post(path, json=v20.Request(method=method_name, params=params).to_json())
    assert raw.status == 200
    assert raw.content_type != 'application/json'
    assert await raw.read() == b''


async def test_errors(json_rpc, path, mocker, aiohttp_client):
    request_id = 1
    params = (1, 2)
    method_name = 'test_method'

    def error_method(*args, **kwargs):
        raise exc.JsonRpcError(code=1, message='message')

    mock = mocker.Mock(name=method_name, side_effect=error_method)

    json_rpc.dispatcher.add(mock, method_name)

    cli = await aiohttp_client(json_rpc.app)
    # method not found
    raw = await cli.post(path, json=v20.Request(method='unknown_method', params=params, id=request_id).to_json())
    assert raw.status == 200

    resp = v20.Response.from_json(await raw.json())
    assert resp.id is request_id
    assert resp.is_error is True
    assert resp.error == exc.MethodNotFoundError(data="method 'unknown_method' not found")

    # customer error
    raw = await cli.post(path, json=v20.Request(method=method_name, params=params, id=request_id).to_json())
    assert raw.status == 200

    resp = v20.Response.from_json(await raw.json())
    mock.assert_called_once_with(args=params)
    assert resp.id == request_id
    assert resp.is_error is True
    assert resp.error == exc.JsonRpcError(code=1, message='message')

    # content type error
    raw = await cli.post(path, data='')
    assert raw.status == 415

    # malformed json
    raw = await cli.post(path, headers={'Content-Type': 'application/json'}, data='')
    assert raw.status == 200
    resp = v20.Response.from_json(await raw.json())
    assert resp.id is None
    assert resp.is_error is True
    assert resp.error == exc.ParseError(data=_)

    # decoding error
    raw = await cli.post(path, headers={'Content-Type': 'application/json'}, data=b'\xff')
    assert raw.status == 400


async def test_context(json_rpc, path, mocker, aiohttp_client):
    request_id = 1
    params = (1, 2)
    method_name = 'test_method'

    # test list parameters
    mock = mocker.Mock(name=method_name, return_value='result')

    json_rpc.dispatcher.add(mock, method_name, context='request')

    cli = await aiohttp_client(json_rpc.app)
    raw = await cli.post(path, json=v20.Request(method=method_name, params=params, id=request_id).to_json())
    assert raw.status == 200

    mock.assert_called_once()
    call_args = mock.call_args[1]
    context, args = call_args['request'], call_args['args']
    assert isinstance(context, web.Request)
    assert args == params

    # test dict parameters
    params = {'param1': 1, 'param2': 2}

    mock.reset_mock()

    cli = await aiohttp_client(json_rpc.app)
    raw = await cli.post(path, json=v20.Request(method=method_name, params=params, id=request_id).to_json())
    assert raw.status == 200

    mock.assert_called_once()
    call_args = mock.call_args[1]
    context, kwargs = call_args['request'], call_args['kwargs']
    assert isinstance(context, web.Request)
    assert kwargs == params



import pytest
from pjrpc.server.validators import pydantic
from pjrpc.server.validators import ValidationError


@pytest.mark.parametrize(
    'dyn_method, params', [
        ('param1: int, param2: int', [1, 2]),
        ('param1: int, *args: int', [1, 2, 3]),
        ('param1: int, *args: int', [1]),
        ('param1: int = 1', []),
        ('param1: int, param2: int = 2', [1]),
        ('param1: int, param2: int', {'param1': 1, 'param2': 2}),
        ('param1: int, **kwargs: int', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1: int, *, param2: int, param3: int', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1: int = 1', {}),
        ('param1: int, param2: int = 2', {'param1': 1}),
        ('param1: int, *, param2: int = 2', {'param1': 1}),
        ('param1, param2: int', [1, 2]),
        ('param1, param2: int', ['param1', 2]),
    ], indirect=['dyn_method'],
)
def test_validation_success(dyn_method, params):
    validator = pydantic.PydanticValidator()
    validator.validate_method(dyn_method, params)
    validator.validate_method(dyn_method, params)


@pytest.mark.parametrize(
    'dyn_method, params', [
        ('param1: int, param2: int', [1]),
        ('param1: int, param2: int', [1, 2, 3]),
        ('param1: int, *args: int', []),
        ('param1: int, param2: int', {'param1': 1}),
        ('param1: int, param2: int', {'param1': 1, 'param2': 2, 'param3': 3}),
        ('param1: int, **kwargs: int', {'param2': 2}),
        ('param1: int, *, param2: int, param3: int', {'param2': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_error(dyn_method, params):
    validator = pydantic.PydanticValidator()

    with pytest.raises(ValidationError):
        validator.validate_method(dyn_method, params)


@pytest.mark.parametrize(
    'dyn_method, exclude, params', [
        ('context, param1: int', ('context',), [1]),
        ('context, *args: int', ('context',), []),
        ('context, param1: int', ('context',), {'param1': 1}),
        ('context, **kwargs: int', ('context',), {}),
        ('context, *, param1: int', ('context',), {'param1': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_exclude_success(dyn_method, exclude, params):
    validator = pydantic.PydanticValidator()
    validator.validate_method(dyn_method, params, exclude=exclude)


@pytest.mark.parametrize(
    'dyn_method, exclude, params', [
        ('context, param1: int', ('context',), [1, 2]),
        ('context, param1: int', ('context',), {'context': '', 'param1': 1}),
        ('context, *, param1: int', ('context',), {'context': 1}),
    ], indirect=['dyn_method'],
)
def test_validation_exclude_error(dyn_method, exclude, params):
    validator = pydantic.PydanticValidator()

    with pytest.raises(ValidationError):
        validator.validate_method(dyn_method, params, exclude=exclude)



import json
import werkzeug
import pytest

from pjrpc import exc
from pjrpc.common import v20
from pjrpc.server.integration import werkzeug as integration

from tests.common import _


@pytest.fixture
def path():
    return '/test/path'


@pytest.fixture
def json_rpc(path):
    return integration.JsonRPC(path)


@pytest.mark.parametrize(
    'request_id, params, result', [
        (
                1,
                (1, 1.1, 'a', {}, False),
                [1, 1.1, 'a', {}, False],
        ),
        (
                'abc',
                {'int': 1, 'float': 1.1, 'str': 'a', 'dict': {}, 'bool': False},
                {'int': 1, 'float': 1.1, 'str': 'a', 'dict': {}, 'bool': False},
        ),
    ],
)
def test_request(json_rpc, path, mocker, request_id, params, result):
    method_name = 'test_method'
    mock = mocker.Mock(name=method_name, return_value=result)

    json_rpc.dispatcher.add(mock, method_name)

    cli = werkzeug.test.Client(json_rpc)
    body_iter, code, header = cli.post(
        path, json=v20.Request(method=method_name, params=params, id=request_id).to_json(),
    )
    body = b''.join(body_iter)
    assert code == '200 OK'

    resp = v20.Response.from_json(json.loads(body))

    if isinstance(params, dict):
        mock.assert_called_once_with(kwargs=params)
    else:
        mock.assert_called_once_with(args=params)

    assert resp.id == request_id
    assert resp.result == result


def test_notify(json_rpc, path, mocker):
    params = [1, 2]
    method_name = 'test_method'
    mock = mocker.Mock(name=method_name, return_value='result')

    json_rpc.dispatcher.add(mock, method_name)

    cli = werkzeug.test.Client(json_rpc)
    body_iter, code, header = cli.post(path, json=v20.Request(method=method_name, params=params).to_json())
    body = b''.join(body_iter)
    assert code == '200 OK'
    assert body == b''


def test_errors(json_rpc, path, mocker):
    request_id = 1
    params = (1, 2)
    method_name = 'test_method'

    def error_method(*args, **kwargs):
        raise exc.JsonRpcError(code=1, message='message')

    mock = mocker.Mock(name=method_name, side_effect=error_method)

    json_rpc.dispatcher.add(mock, method_name)

    cli = werkzeug.test.Client(json_rpc)
    # method not found
    body_iter, code, header = cli.post(
        path, json=v20.Request(method='unknown_method', params=params, id=request_id).to_json(),
    )
    body = b''.join(body_iter)
    assert code == '200 OK'

    resp = v20.Response.from_json(json.loads(body))
    assert resp.id is request_id
    assert resp.is_error is True
    assert resp.error == exc.MethodNotFoundError(data="method 'unknown_method' not found")

    # customer error
    body_iter, code, header = cli.post(
        path, json=v20.Request(method=method_name, params=params, id=request_id).to_json(),
    )
    body = b''.join(body_iter)
    assert code == '200 OK'

    resp = v20.Response.from_json(json.loads(body))
    mock.assert_called_once_with(args=params)
    assert resp.id == request_id
    assert resp.is_error is True
    assert resp.error == exc.JsonRpcError(code=1, message='message')

    # malformed json
    body_iter, code, header = cli.post(path, headers={'Content-Type': 'application/json'}, data='')
    body = b''.join(body_iter)
    assert code == '200 OK'
    resp = v20.Response.from_json(json.loads(body))
    assert resp.id is None
    assert resp.is_error is True
    assert resp.error == exc.ParseError(data=_)



import pytest

import pjrpc


def test_error_serialization():
    error = pjrpc.exc.ServerError()

    actual_dict = error.to_json()
    expected_dict = {
        'code': -32000,
        'message': 'Server error',
    }

    assert actual_dict == expected_dict


def test_error_deserialization():
    data = {
        'code': -32000,
        'message': 'Server error',
    }

    error = pjrpc.exc.ServerError.from_json(data)

    assert error.code == -32000
    assert error.message == 'Server error'


def test_error_data_serialization():
    error = pjrpc.exc.MethodNotFoundError(data='method_name')

    actual_dict = error.to_json()
    expected_dict = {
        'code': -32601,
        'message': 'Method not found',
        'data': 'method_name',
    }

    assert actual_dict == expected_dict


def test_custom_error_data_serialization():
    error = pjrpc.exc.JsonRpcError(code=2001, message='Custom error', data='additional data')

    actual_dict = error.to_json()
    expected_dict = {
        'code': 2001,
        'message': 'Custom error',
        'data': 'additional data',
    }

    assert actual_dict == expected_dict


def test_custom_error_data_deserialization():
    data = {
        'code': -32601,
        'message': 'Method not found',
        'data': 'method_name',
    }

    error = pjrpc.exc.JsonRpcError.from_json(data)

    assert error.code == -32601
    assert error.message == 'Method not found'
    assert error.data == 'method_name'


def test_error_deserialization_errors():
    with pytest.raises(pjrpc.exc.DeserializationError, match="data must be of type dict"):
        pjrpc.exc.JsonRpcError.from_json([])

    with pytest.raises(pjrpc.exc.DeserializationError, match="required field 'message' not found"):
        pjrpc.exc.JsonRpcError.from_json({'code': 1})

    with pytest.raises(pjrpc.exc.DeserializationError, match="required field 'code' not found"):
        pjrpc.exc.JsonRpcError.from_json({'message': ""})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'code' must be of type integer"):
        pjrpc.exc.JsonRpcError.from_json({'code': "1", 'message': ""})

    with pytest.raises(pjrpc.exc.DeserializationError, match="field 'message' must be of type string"):
        pjrpc.exc.JsonRpcError.from_json({'code': 1, 'message': 2})


def test_error_repr():
    assert repr(pjrpc.exc.ServerError(data='data')) == "ServerError(code=-32000, message='Server error', data='data')"
    assert str(pjrpc.exc.ServerError()) == "(-32000) Server error"


def test_custom_error_registration():
    data = {
        'code': 2000,
        'message': 'Custom error',
        'data': 'custom_data',
    }

    class CustomError(pjrpc.exc.JsonRpcError):
        code = 2000
        message = 'Custom error'

    error = pjrpc.exc.JsonRpcError.from_json(data)

    assert isinstance(error, CustomError)
    assert error.code == 2000
    assert error.message == 'Custom error'
    assert error.data == 'custom_data'



import inspect
from typing import Any, Callable, Dict, Iterable, Optional

import pydantic as pd

from pjrpc.common import UNSET
from pjrpc.server.specs.extractors import BaseSchemaExtractor, Schema


class PydanticSchemaExtractor(BaseSchemaExtractor):
    """
    Pydantic method specification extractor.
    """

    def extract_params_schema(self, method: Callable, exclude: Iterable[str] = ()) -> Dict[str, Schema]:
        exclude = set(exclude)
        signature = inspect.signature(method)

        field_definitions = {}
        for param in signature.parameters.values():
            if param.name in exclude:
                continue

            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        params_model = pd.create_model('RequestModel', **field_definitions)
        model_schema = params_model.schema(ref_template='{model}')

        parameters_schema = {}
        for param_name, param_schema in model_schema['properties'].items():
            param_schema = self._extract_field_schema(model_schema, field_name=param_name)
            required = param_name in model_schema.get('required', [])

            parameters_schema[param_name] = Schema(
                schema=param_schema,
                summary=param_schema.get('title', UNSET),
                description=param_schema.get('description', UNSET),
                deprecated=param_schema.get('deprecated', UNSET),
                required=required,
            )

        return parameters_schema

    def extract_result_schema(self, method: Callable) -> Schema:
        result = inspect.signature(method)

        if result.return_annotation is inspect.Parameter.empty:
            return_annotation = Any
        elif result.return_annotation is None:
            return_annotation = Optional[None]
        else:
            return_annotation = result.return_annotation

        result_model = pd.create_model('ResultModel', result=(return_annotation, pd.fields.Undefined))
        model_schema = result_model.schema(ref_template='{model}')

        result_schema = self._extract_field_schema(model_schema, field_name='result')
        required = 'result' in model_schema.get('required', [])
        if not required:
            result_schema['nullable'] = 'true'

        result_schema = Schema(
            schema=result_schema,
            summary=result_schema.get('title', UNSET),
            description=result_schema.get('description', UNSET),
            deprecated=result_schema.get('deprecated', UNSET),
            required=required,
        )

        return result_schema

    @staticmethod
    def _extract_field_schema(model_schema: Dict[str, Any], field_name: str) -> Dict[str, Any]:
        field_schema = model_schema['properties'][field_name]
        if '$ref' in field_schema:
            field_schema = model_schema['definitions'][field_schema['$ref']]

        return field_schema



import json

import pytest
import responses
import respx

import pjrpc
from pjrpc.client.backend import requests as requests_backend
from pjrpc.client.backend import httpx as httpx_backend


class ResponsesMocker(responses.RequestsMock):

    def mock(self, method=None, url=None, status=None, content=None, json=None):
        self.add(method=method, url=url, status=status, body=content, json=json)

    @property
    def requests(self):
        requests = []
        for request, response in self.calls:
            request.content = request.body
            requests.append(request)

        return requests


class RespxMocker(respx.MockRouter):

    def mock(self, method=None, url=None, status=None, content=None, json=None):
        route = self.route(method=method, url=url)
        route.respond(status_code=status, content=content, json=json)

    @property
    def requests(self):
        return [request for request, response in self.calls]


@pytest.mark.parametrize(
    'Client, mocker', [
        (requests_backend.Client, ResponsesMocker),
        (httpx_backend.Client, RespxMocker),
    ],
)
def test_call(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        client = Client(test_url)

        response = client.send(pjrpc.Request('method', (1, 2), id=1))
        assert response.id == 1
        assert response.result == 'result'

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = client.call('method', arg1=1, arg2=2)
        assert result == 'result'

        assert mock.requests[1].url == test_url
        assert json.loads(mock.requests[1].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': {'arg1': 1, 'arg2': 2},
        }

        result = client('method', 1, 2)
        assert result == 'result'

        assert mock.requests[2].url == test_url
        assert json.loads(mock.requests[2].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = client.proxy.method(1, 2)
        assert result == 'result'

        assert mock.requests[3].url == test_url
        assert json.loads(mock.requests[3].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

        result = client.proxy.method(arg1=1, arg2=2)
        assert result == 'result'

        assert mock.requests[4].url == test_url
        assert json.loads(mock.requests[4].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': {'arg1': 1, 'arg2': 2},
        }


@pytest.mark.parametrize(
    'Client, mocker', [
        (requests_backend.Client, ResponsesMocker),
        (httpx_backend.Client, RespxMocker),
    ],
)
def test_notify(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock('POST', test_url, status=200, content='')
        client = Client(test_url)

        response = client.send(pjrpc.Request('method', params=[1, 2]))
        assert response is None
        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': [1, 2],
        }

        response = client.notify('method', 1, 2)
        assert response is None
        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'method': 'method',
            'params': [1, 2],
        }


@pytest.mark.parametrize(
    'Client, mocker', [
        (requests_backend.Client, ResponsesMocker),
        (httpx_backend.Client, RespxMocker),
    ],
)
def test_batch(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json=[
                {
                    'jsonrpc': '2.0',
                    'id': 1,
                    'result': 'result1',
                },
                {
                    'jsonrpc': '2.0',
                    'id': 2,
                    'result': 2,
                },
            ],
        )

        client = Client(test_url)

        result = client.batch.send(
            pjrpc.BatchRequest(
                pjrpc.Request('method1', params=[1, 2], id=1),
                pjrpc.Request('method2', params=[2, 3], id=2),
                pjrpc.Request('method3', params=[3, 4]),
            ),
        )
        assert len(result) == 2
        assert result[0].id == 1
        assert result[0].result == 'result1'
        assert result[1].id == 2
        assert result[1].result == 2

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method3',
                'params': [3, 4],
            },
        ]

        result = client.batch[
            ('method1', 1, 2),
            ('method2', 2, 3),
        ]
        assert result == ('result1', 2)

        assert mock.requests[1].url == test_url
        assert json.loads(mock.requests[1].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = client.batch('method1', 1, 2)('method2', 2, 3).call()
        assert result == ('result1', 2)

        assert mock.requests[2].url == test_url
        assert json.loads(mock.requests[2].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = client.batch.proxy.method1(1, 2).method2(2, 3)()
        assert result == ('result1', 2)

        assert mock.requests[3].url == test_url
        assert json.loads(mock.requests[3].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

        result = client.batch.send(
            pjrpc.BatchRequest(
                pjrpc.Request(method='method1', params=[1, 2], id=1),
                pjrpc.Request(method='method2', params=[2, 3], id=2),
            ),
        )
        assert result[0].id == 1
        assert result[0].result == 'result1'
        assert result[1].id == 2
        assert result[1].result == 2

        assert mock.requests[4].url == test_url
        assert json.loads(mock.requests[4].content) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3],
            },
        ]

    with mocker() as mock:
        mock.mock('POST', test_url, status=200)
        result = client.batch.notify('method1', 1, 2).notify('method2', 2, 3).call()
        assert result is None

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == [
            {
                'jsonrpc': '2.0',
                'method': 'method1',
                'params': [1, 2],
            },
            {
                'jsonrpc': '2.0',
                'method': 'method2',
                'params': [2, 3],
            },
        ]


@pytest.mark.parametrize(
    'Client, mocker', [
        (requests_backend.Client, ResponsesMocker),
        (httpx_backend.Client, RespxMocker),
    ],
)
def test_error(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'id': 1,
                'error': {
                    'code': -32601,
                    'message': 'Method not found',
                },
            },
        )

        client = Client(test_url)

        with pytest.raises(pjrpc.exceptions.MethodNotFoundError):
            client('method', 1, 2)

        assert mock.requests[0].url == test_url
        assert json.loads(mock.requests[0].content) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2],
        }

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'error': {
                    'code': -32600,
                    'message': 'Invalid request',
                },
            },
        )

        with pytest.raises(pjrpc.exceptions.InvalidRequestError):
            client.batch[('method', 'param')]

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json=[
                {
                    'jsonrpc': '2.0',
                    'id': 1,
                    'result': 'result1',
                },
                {
                    'jsonrpc': '2.0',
                    'id': 3,
                    'result': 2,
                },
            ],
        )

        client = Client(test_url)

        with pytest.raises(pjrpc.exceptions.IdentityError):
            client.batch[
                ('method1', 'param'),
                ('method2', 'param'),
            ]


@pytest.mark.parametrize(
    'Client, mocker', [
        (requests_backend.Client, ResponsesMocker),
        (httpx_backend.Client, RespxMocker),
    ],
)
def test_context_manager(Client, mocker):
    test_url = 'http://test.com/api'

    with mocker() as mock:
        mock.mock(
            'POST', test_url, status=200, json={
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result',
            },
        )

        with Client(test_url) as client:
            response = client.send(pjrpc.Request('method', (1, 2), id=1))
            assert response.id == 1
            assert response.result == 'result'



"""
OpenRPC specification generator. See https://spec.open-rpc.org/.
"""
import enum

try:
    import dataclasses as dc
except ImportError:
    raise AssertionError("python 3.7 or later is required")

from typing import Any, Callable, Dict, Iterable, List, Optional, Type, Union

from pjrpc.common import exceptions, UNSET
from pjrpc.server import utils, Method

from . import extractors
from . import Specification

Json = Union[str, int, float, dict, bool, list, tuple, set, None]


@dc.dataclass(frozen=True)
class Contact:
    """
    Contact information for the exposed API.

    :param name: the identifying name of the contact person/organization
    :param url: the URL pointing to the contact information
    :param email: the email address of the contact person/organization
    """

    name: str = UNSET
    url: str = UNSET
    email: str = UNSET


@dc.dataclass(frozen=True)
class License:
    """
    License information for the exposed API.

    :param name: the license name used for the API
    :param url: a URL to the license used for the API
    """

    name: str
    url: str = UNSET


@dc.dataclass(frozen=True)
class Info:
    """
    Metadata about the API.

    :param title: the title of the application
    :param version: the version of the OpenRPC document
    :param description: a verbose description of the application
    :param contact: the contact information for the exposed API
    :param license: the license information for the exposed API
    :param termsOfService: a URL to the Terms of Service for the API
    """

    title: str
    version: str
    description: str = UNSET
    contact: Contact = UNSET
    license: License = UNSET
    termsOfService: str = UNSET


@dc.dataclass(frozen=True)
class Server:
    """
    Connectivity information of a target server.

    :param name: a name to be used as the canonical name for the server.
    :param url: a URL to the target host
    :param summary: a short summary of what the server is
    :param description: an optional string describing the host designated by the URL
    """

    name: str
    url: str
    summary: str = UNSET
    description: str = UNSET


@dc.dataclass(frozen=True)
class ExternalDocumentation:
    """
    Allows referencing an external resource for extended documentation.

    :param url: A verbose explanation of the target documentation
    :param description: The URL for the target documentation. Value MUST be in the format of a URL
    """

    url: str
    description: str = UNSET


@dc.dataclass(frozen=True)
class Tag:
    """
    A list of tags for API documentation control.
    Tags can be used for logical grouping of methods by resources or any other qualifier.

    :param name: the name of the tag
    :param summary: a short summary of the tag
    :param description: a verbose explanation for the tag
    :param externalDocs: additional external documentation for this tag
    """

    name: str
    summary: str = UNSET
    description: str = UNSET
    externalDocs: ExternalDocumentation = UNSET


@dc.dataclass(frozen=True)
class ExampleObject:
    """
    The ExampleObject object is an object the defines an example.

    :param value: embedded literal example
    :param name: canonical name of the example
    :param summary: short description for the example
    :param description: a verbose explanation of the example
    """

    value: Json
    name: str
    summary: str = UNSET
    description: str = UNSET


@dc.dataclass(frozen=True)
class MethodExample:
    """
    The example Pairing object consists of a set of example params and result.

    :param params: example parameters
    :param result: example result
    :param name: name for the example pairing
    :param summary: short description for the example pairing
    :param description: a verbose explanation of the example pairing
    """

    name: str
    params: List[ExampleObject]
    result: ExampleObject
    summary: str = UNSET
    description: str = UNSET


@dc.dataclass(frozen=True)
class ContentDescriptor:
    """
    Content Descriptors are objects that describe content.
    They are reusable ways of describing either parameters or result.

    :param name: name of the content that is being described
    :param schema: schema that describes the content. The Schema Objects MUST follow the specifications outline
                   in the JSON Schema Specification 7 (https://json-schema.org/draft-07/json-schema-release-notes.html)
    :param summary: a short summary of the content that is being described
    :param description: a verbose explanation of the content descriptor behavior
    :param required: determines if the content is a required field
    :param deprecated: specifies that the content is deprecated and SHOULD be transitioned out of usage
    """

    name: str
    schema: Dict[str, Any]
    summary: str = UNSET
    description: str = UNSET
    required: bool = UNSET
    deprecated: bool = UNSET


@dc.dataclass(frozen=True)
class Error:
    """
    Defines an application level error.

    :param code: a Number that indicates the error type that occurred
    :param message: a String providing a short description of the error
    :param data: a Primitive or Structured value that contains additional information about the error
    """

    code: int
    message: str
    data: Dict[str, Any] = UNSET


class ParamStructure(str, enum.Enum):
    """
    The expected format of the parameters.
    """

    BY_NAME = 'by-name'
    BY_POSITION = 'by-position'
    EITHER = 'either'


@dc.dataclass(frozen=True)
class MethodInfo:
    """
    Describes the interface for the given method name.

    :param name: the canonical name for the method
    :param params: a list of parameters that are applicable for this method
    :param result: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage examples
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    :param paramStructure: the expected format of the parameters
    :param externalDocs: additional external documentation for this method
    :param servers: an alternative servers array to service this method
    """

    name: str
    params: List[Union[ContentDescriptor, dict]]
    result: Union[ContentDescriptor, dict]
    errors: List[Error] = UNSET
    paramStructure: ParamStructure = UNSET
    examples: List[MethodExample] = UNSET
    summary: str = UNSET
    description: str = UNSET
    tags: List[Tag] = UNSET
    deprecated: bool = UNSET
    externalDocs: ExternalDocumentation = UNSET
    servers: List[Server] = UNSET


def annotate(
    params_schema: List[ContentDescriptor] = UNSET,
    result_schema: ContentDescriptor = UNSET,
    errors: List[Union[Error, Type[exceptions.JsonRpcError]]] = UNSET,
    examples: List[MethodExample] = UNSET,
    summary: str = UNSET,
    description: str = UNSET,
    tags: List[Union[Tag, str]] = UNSET,
    deprecated: bool = UNSET,
):
    """
    Adds JSON-RPC method to the API specification.

    :param params_schema: a list of parameters that are applicable for this method
    :param result_schema: the description of the result returned by the method
    :param errors: a list of custom application defined errors that MAY be returned
    :param examples: method usage example
    :param summary: a short summary of what the method does
    :param description: a verbose explanation of the method behavior
    :param tags: a list of tags for API documentation control
    :param deprecated: declares this method to be deprecated
    """

    def decorator(method: Callable) -> Callable:
        utils.set_meta(
            method,
            openrpc_spec=dict(
                params_schema=params_schema,
                result_schema=result_schema,
                errors=[
                    error if isinstance(error, Error) else Error(code=error.code, message=error.message)
                    for error in errors
                ] if errors else UNSET,
                examples=examples,
                tags=[
                    Tag(name=tag) if isinstance(tag, str) else tag
                    for tag in tags
                ] if tags else UNSET,
                summary=summary,
                description=description,
                deprecated=deprecated,
            ),
        )

        return method

    return decorator


@dc.dataclass(init=False)
class OpenRPC(Specification):
    """
    OpenRPC Specification.

    :param info: specification information
    :param path: specification url path
    :param servers: connectivity information
    :param external_docs: additional external documentation
    :param openrpc: the semantic version number of the OpenRPC Specification version that the OpenRPC document uses
    :param schema_extractor: method specification extractor
    """

    info: Info
    methods: List[MethodInfo] = dc.field(default_factory=list)
    servers: List[Server] = UNSET
    externalDocs: ExternalDocumentation = UNSET
    openrpc: str = '1.0.0'

    def __init__(
        self,
        info: Info,
        path: str = '/openrpc.json',
        servers: List[Server] = UNSET,
        external_docs: Optional[ExternalDocumentation] = UNSET,
        openrpc: str = '1.0.0',
        schema_extractor: Optional[extractors.BaseSchemaExtractor] = None,
    ):
        super().__init__(path)

        self.info = info
        self.servers = servers
        self.externalDocs = external_docs
        self.openrpc = openrpc
        self.methods = []

        self._schema_extractor = schema_extractor

    def schema(self, path: str, methods: Iterable[Method]) -> dict:
        for method in methods:
            method_name = method.name
            path = path.rstrip('/')

            method_meta = utils.get_meta(method.method)
            annotated_spec = method_meta.get('openrpc_spec', {})

            params_schema = self._schema_extractor.extract_params_schema(method.method, exclude=[method.context])
            result_schema = self._schema_extractor.extract_result_schema(method.method)
            extracted_spec = dict(
                params_schema=[
                    ContentDescriptor(
                        name=name,
                        schema=schema.schema,
                        summary=schema.summary,
                        description=schema.description,
                        required=schema.required,
                        deprecated=schema.deprecated,
                    ) for name, schema in params_schema.items()
                ],
                result_schema=ContentDescriptor(
                    name='result',
                    schema=result_schema.schema,
                    summary=result_schema.summary,
                    description=result_schema.description,
                    required=result_schema.required,
                    deprecated=result_schema.deprecated,
                ),
                errors=self._schema_extractor.extract_errors_schema(method.method),
                deprecated=self._schema_extractor.extract_deprecation_status(method.method),
                description=self._schema_extractor.extract_description(method.method),
                summary=self._schema_extractor.extract_summary(method.method),
                tags=self._schema_extractor.extract_tags(method.method),
                examples=[
                    MethodExample(
                        name=example.summary,
                        params=[
                            ExampleObject(
                                value=param_value,
                                name=param_name,
                            )
                            for param_name, param_value in example.params.items()
                        ],
                        result=ExampleObject(
                            name='result',
                            value=example.result,
                        ),
                        summary=example.summary,
                        description=example.description,
                    )
                    for example in self._schema_extractor.extract_examples(method.method) or []
                ],
            )
            method_spec = extracted_spec.copy()
            method_spec.update((k, v) for k, v in annotated_spec.items() if v is not UNSET)

            self.methods.append(
                MethodInfo(
                    name=method_name,
                    params=method_spec['params_schema'],
                    result=method_spec['result_schema'],
                    errors=method_spec['errors'],
                    examples=method_spec['examples'],
                    summary=method_spec['summary'],
                    description=method_spec['description'],
                    tags=method_spec['tags'],
                    deprecated=method_spec['deprecated'],
                ),
            )

        return dc.asdict(
            self,
            dict_factory=lambda iterable: dict(
                filter(lambda item: item[1] is not UNSET, iterable),
            ),
        )



import dataclasses as dc
import inspect
import itertools as it
from typing import Any, Callable, Dict, Iterable, List, Union

from pjrpc.common import UNSET, UnsetType


@dc.dataclass(frozen=True)
class Schema:
    """
    Method parameter/result schema.
    """

    schema: Dict[str, Any]
    required: bool = True
    summary: str = UNSET
    description: str = UNSET
    deprecated: bool = UNSET


@dc.dataclass(frozen=True)
class Example:
    """
    Method usage example.
    """

    params: Dict[str, Any]
    result: Any
    version: str = '2.0'
    summary: str = UNSET
    description: str = UNSET


@dc.dataclass(frozen=True)
class Tag:
    """
    A list of method tags.
    """

    name: str
    description: str = UNSET
    externalDocs: str = UNSET


@dc.dataclass(frozen=True)
class Error:
    """
    Defines an application level error.
    """

    code: int
    message: str
    data: Dict[str, Any] = UNSET


class BaseSchemaExtractor:
    """
    Base method schema extractor.
    """

    def extract_params_schema(self, method: Callable, exclude: Iterable[str] = ()) -> Dict[str, Schema]:
        """
        Extracts method parameters schema.
        """

        return {}

    def extract_result_schema(self, method: Callable) -> Schema:
        """
        Extracts method result schema.
        """

        return Schema(schema={})

    def extract_description(self, method: Callable) -> Union[UnsetType, str]:
        """
        Extracts method description.
        """

        if method.__doc__:
            doc = inspect.cleandoc(method.__doc__)
            description = '\n'.join(it.takewhile(lambda line: line, doc.split('\n')))
        else:
            description = UNSET

        return description

    def extract_summary(self, method: Callable) -> Union[UnsetType, str]:
        """
        Extracts method summary.
        """

        description = self.extract_description(method)
        if description:
            summary = description.split('.')[0]
        else:
            summary = UNSET

        return summary

    def extract_errors_schema(self, method: Callable) -> Union[UnsetType, List[Error]]:
        """
        Extracts method errors schema.
        """

        return UNSET

    def extract_tags(self, method: Callable) -> Union[UnsetType, List[Tag]]:
        """
        Extracts method tags.
        """

        return UNSET

    def extract_examples(self, method: Callable) -> Union[UnsetType, List[Example]]:
        """
        Extracts method usage examples.
        """

        return UNSET

    def extract_deprecation_status(self, method: Callable) -> Union[UnsetType, bool]:
        """
        Extracts method deprecation status.
        """

        return UNSET



#!/usr/bin/env python

import pathlib
import setuptools.command.test
import sys
from setuptools import setup, find_packages

requirements = [
]

test_requirements = [
    'aioresponses~=0.0',
    'pytest~=6.0',
    'pytest-aiohttp~=0.0',
    'pytest-mock~=1.0',
    'responses~=0.0',
]

with open('README.rst', 'r') as file:
    readme = file.read()


def parse_about():
    about_globals = {}
    this_path = pathlib.Path(__file__).parent
    about_module_text = pathlib.Path(this_path, 'pjrpc', '__about__.py').read_text()
    exec(about_module_text, about_globals)

    return about_globals


about = parse_about()


class PyTest(setuptools.command.test.test):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to py.test')]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.pytest_args))


setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__url__'],
    license=about['__license__'],
    keywords=[
        'json-rpc', 'rpc', 'jsonrpc-client', 'jsonrpc-server',
        'requests', 'aiohttp', 'flask', 'httpx', 'aio-pika', 'kombu',
        'openapi', 'openrpc',
    ],
    python_requires=">=3.6",
    packages=find_packages(),
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require={
        'aiohttp': ['aiohttp~=3.0'],
        'aio-pika': ['aio-pika~=6.0'],
        'flask': ['flask~=1.0'],
        'jsonschema': ['jsonschema~=3.0'],
        'kombu': ['kombu~=5.0'],
        'pydantic': ['pydantic~=1.0'],
        'requests': ['requests~=2.0'],
        'httpx': ['requests~=0.0'],
        'docstring-parser': ['docstring-parser~=0.0'],
        'openapi-ui-bundles': ['openapi-ui-bundles~=0.0'],
    },
    entry_points={"pytest11": ["pjrpc = pjrpc.client.integrations.pytest"]},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: Public Domain',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    project_urls={
        "Documentation": "https://pjrpc.readthedocs.io/en/latest/",
        'Source': 'https://github.com/dapper91/pjrpc',
    },
    cmdclass={'test': PyTest},
)



import json

import pjrpc
from pjrpc.server import dispatcher as pjrpc_dispatcher


def test_error_handlers(mocker):
    test_request = pjrpc.common.Request('test_method', params=dict(param='param'), id=1)
    test_context = object()
    expected_error = pjrpc.exceptions.MethodNotFoundError(data="method 'test_method' not found")
    expected_response = pjrpc.common.Response(id=1, error=expected_error)
    handler_call_order = []

    def test_any_error_handler(request, context, error):
        handler_call_order.append(test_any_error_handler)
        assert request == test_request
        assert error == expected_error
        assert context is test_context

        return error

    def test_32601_error_handler(request, context, error):
        handler_call_order.append(test_32601_error_handler)
        assert request == test_request
        assert error == expected_error
        assert context is test_context

        return error

    def test_32603_error_handler(request, context, error):
        assert False, 'should not be called'

    dispatcher = pjrpc_dispatcher.Dispatcher(
        error_handlers={
            None: [test_any_error_handler],
            -32601: [test_32601_error_handler],
            -32603: [test_32603_error_handler],
        },
    )

    request_text = json.dumps(test_request.to_json())
    response_text = dispatcher.dispatch(request_text, test_context)
    actual_response = pjrpc.common.Response.from_json(json.loads(response_text))
    assert actual_response == expected_response

    assert handler_call_order == [test_any_error_handler, test_32601_error_handler]


async def test_async_error_handlers(mocker):
    test_request = pjrpc.common.Request('test_method', params=dict(param='param'), id=1)
    test_context = object()
    expected_error = pjrpc.exceptions.MethodNotFoundError(data="method 'test_method' not found")
    expected_response = pjrpc.common.Response(id=1, error=expected_error)
    handler_call_order = []

    async def test_any_error_handler(request, context, error):
        handler_call_order.append(test_any_error_handler)
        assert request == test_request
        assert error == expected_error
        assert context is test_context

        return error

    async def test_32601_error_handler(request, context, error):
        handler_call_order.append(test_32601_error_handler)
        assert request == test_request
        assert error == expected_error
        assert context is test_context

        return error

    async def test_32603_error_handler(request, context, error):
        assert False, 'should not be called'

    dispatcher = pjrpc_dispatcher.AsyncDispatcher(
        error_handlers={
            None: [test_any_error_handler],
            -32601: [test_32601_error_handler],
            -32603: [test_32603_error_handler],
        },
    )

    request_text = json.dumps(test_request.to_json())
    response_text = await dispatcher.dispatch(request_text, test_context)
    actual_response = pjrpc.common.Response.from_json(json.loads(response_text))
    assert actual_response == expected_response

    assert handler_call_order == [test_any_error_handler, test_32601_error_handler]



"""
Client and server common functions, types and classes that implements JSON-RPC protocol itself
and agnostic to any transport protocol layer (http, socket, amqp) and server-side implementation.
"""

from . import generators
from .common import UNSET, UnsetType, JSONEncoder
from .v20 import Request, Response, BatchRequest, BatchResponse


__all__ = [
    'Request',
    'Response',
    'BatchRequest',
    'BatchResponse',
    'UNSET',
    'UnsetType',
    'JSONEncoder',
    'generators',
]



import random
import string
import uuid

import pytest
from pjrpc.common import generators


@pytest.mark.parametrize(
    'start, step, length, result', [
        (0, 1, 3, [0, 1, 2]),
        (1, 2, 3, [1, 3, 5]),
    ],
)
def test_sequential(start, step, length, result):
    gen = generators.sequential(start, step)
    assert [next(gen) for _ in range(length)] == result


@pytest.mark.parametrize(
    'a, b, length, seed, result', [
        (0, 10, 5, 1, [2, 9, 1, 4, 1]),
    ],
)
def test_randint(a, b, length, seed, result):
    random.seed(seed)
    gen = generators.randint(a, b)
    assert [next(gen) for _ in range(length)] == result


@pytest.mark.parametrize(
    'length, chars, seed, result', [
        (5, string.ascii_lowercase, 1, ['eszyc', 'idpyo', 'pumzg', 'dpamn', 'tyyaw']),
        (5, string.digits, 1, ['29141', '77763', '17066', '90743', '91500']),
    ],
)
def test_random(length, chars, seed, result):
    random.seed(seed)
    gen = generators.random(length, chars)
    assert [next(gen) for _ in range(length)] == result


def test_uuid(mocker):
    mocked_uuid = uuid.UUID('226a2c23-c98b-4729-b398-0dae550e99ff')
    mocker.patch('uuid.uuid4', return_value=mocked_uuid)

    gen = generators.uuid()
    assert [next(gen) for _ in range(2)] == [mocked_uuid] * 2



import pathlib
import pytest
from typing import Callable, Optional

THIS_DIR = pathlib.Path(__file__).parent


@pytest.fixture
def dyn_method(request):
    signature = request.param
    context = globals().copy()
    exec(f"def dynamic_method({signature}): pass", context)

    return context['dynamic_method']


@pytest.fixture
def resources():
    def getter(name: str, loader: Optional[Callable] = None) -> str:
        resource_file = THIS_DIR / 'resources' / name
        data = resource_file.read_text()
        if loader:
            return loader(data)
        else:
            return data

    return getter
